import 'dart:async';
import 'dart:typed_data';

import 'package:flutter/material.dart';
import 'package:forme/forme.dart';
import 'package:image_picker/image_picker.dart';

import 'draggable_grid_view.dart';

typedef ImagePickerBuilder = Widget Function(
  ValueChanged<List<FormeImage>> insertPickedImages,
);

typedef GridItemBuilder = Widget Function(
  ValueChanged<int> remove,
  FormeImage item,
  int index,
);

class FormeImagePicker extends FormeField<List<FormeImage>> {
  final int? maximum;
  final DraggableConfiguration<FormeImage>? draggableConfiguration;
  FormeImagePicker({
    Duration? animateDuration,
    this.maximum,
    this.draggableConfiguration,
    ImagePickerBuilder? pickerBuilder,
    GridItemBuilder? gridItemBuilder,
    EdgeInsetsGeometry? gridItemPadding,
    required SliverGridDelegate gridDelegate,
    List<FormeImage>? initialValue,
    required String name,
    bool readOnly = false,
    Key? key,
    InputDecoration? decoration,
    FormeFieldDecorator<List<FormeImage>>? decorator,
    int? order,
    bool quietlyValidate = false,
    Duration? asyncValidatorDebounce,
    AutovalidateMode? autovalidateMode,
    FormeValueChanged<List<FormeImage>>? onValueChanged,
    FormeFocusChanged<List<FormeImage>>? onFocusChanged,
    FormeFieldValidationChanged<List<FormeImage>>? onValidationChanged,
    FormeFieldInitialed<List<FormeImage>>? onInitialed,
    FormeFieldSetter<List<FormeImage>>? onSaved,
    FormeValidator<List<FormeImage>>? validator,
    FormeAsyncValidator<List<FormeImage>>? asyncValidator,
    bool requestFocusOnUserInteraction = true,
    bool registrable = true,
    bool enabled = true,
  }) : super(
            enabled: enabled,
            registrable: registrable,
            requestFocusOnUserInteraction: requestFocusOnUserInteraction,
            order: order,
            quietlyValidate: quietlyValidate,
            asyncValidatorDebounce: asyncValidatorDebounce,
            autovalidateMode: autovalidateMode,
            onValueChanged: onValueChanged,
            onFocusChanged: onFocusChanged,
            onValidationChanged: onValidationChanged,
            onInitialed: onInitialed,
            onSaved: onSaved,
            validator: validator,
            asyncValidator: asyncValidator,
            key: key,
            decorator: decorator ??
                (decoration == null
                    ? null
                    : FormeInputDecoratorBuilder(decoration: decoration)),
            readOnly: readOnly,
            name: name,
            initialValue: initialValue ?? <FormeImage>[],
            builder: (genericState) {
              final _FormeImagePickerState state =
                  genericState as _FormeImagePickerState;
              return DraggableGridView<_Item<FormeImage>>(
                animateDuration:
                    animateDuration ?? const Duration(milliseconds: 150),
                gridDelegate: gridDelegate,
                onAccept: state._onAccept,
                builder: (context, item, index) {
                  if (maximum != null &&
                      state.gridController.value.length - 1 >= maximum &&
                      item.value == null) {
                    return Container();
                  }
                  Widget child;
                  if (item.value == null) {
                    child = (pickerBuilder ?? state._defaultImagePicker)
                        .call(state._insertPickedImages);
                  } else {
                    child = (gridItemBuilder ?? state._defaultGridItem)
                        .call(state._remove, item.value!, index);
                  }
                  return Padding(
                    padding: gridItemPadding ?? const EdgeInsets.all(5),
                    child: child,
                  );
                },
                draggableConfiguration: state._draggableConfiguration(),
                controller: state.gridController,
              );
            });

  @override
  FormeFieldState<List<FormeImage>> createState() => _FormeImagePickerState();
}

class _FormeImagePickerState extends FormeFieldState<List<FormeImage>> {
  @override
  FormeImagePicker get widget => super.widget as FormeImagePicker;

  late GridController<_Item<FormeImage>> gridController;

  @override
  void beforeInitiation() {
    super.beforeInitiation();
    gridController = GridController(_convert(List.of(initialValue)));
  }

  @override
  void updateFieldValueInDidUpdateWidget(
      FormeField<List<FormeImage>> oldWidget) {
    if (widget.maximum != null) {
      if (widget.maximum! <= value.length) {
        setValue(List.of(value).sublist(0, widget.maximum));
      } else {
        gridController.value = List.of(gridController.value)
            .where((element) => element.value != null)
            .toList()
          ..add(_Item.empty());
      }
    }
  }

  List<_Item<FormeImage>> _convert(List<FormeImage> images) {
    final List<_Item<FormeImage>> items = images.map((e) => _Item(e)).toList();
    if (widget.maximum == null || items.length < widget.maximum!) {
      items.add(_Item.empty());
    }
    return items;
  }

  @override
  void afterInitiation() {
    super.afterInitiation();
    gridController.addListener(() {
      didChange(gridController.value
          .where((element) => element.value != null)
          .map((e) => e.value!)
          .toList());
    });
  }

  @override
  void dispose() {
    gridController.dispose();
    super.dispose();
  }

  @override
  void onValueChanged(List<FormeImage> newValue) {
    final List<FormeImage> gridValue = gridController.value
        .where((element) => element.value != null)
        .map((e) => e.value!)
        .toList();
    if (!compareValue(gridValue, newValue)) {
      gridController.value = _convert(List.of(newValue));
    }
  }

  /// this method will decode image first rather than directly return [ImageProvider]
  Future<ImageProvider> _load(FormeImage file,
      [ImageConfiguration? configuration]) async {
    final ImageProvider image = await file.image;
    final Completer<ImageProvider> completer = Completer();
    final ImageStream stream =
        image.resolve(configuration ?? const ImageConfiguration());
    final ImageStreamListener listener =
        ImageStreamListener((info, synchronousCall) {
      file._broken = false;
      file._cache = image;
      completer.complete(image);
    }, onError: (error, trace) {
      file._broken = true;
      file._cache = null;
      completer.completeError(error, trace);
    });
    stream.addListener(listener);
    final Future<ImageProvider> future = completer.future;
    return future.whenComplete(() {
      //https://github.com/Sub6Resources/flutter_html/issues/780
      stream.removeListener(listener);
    });
  }

  void _remove(int index) {
    if (!mounted) {
      return;
    }
    gridController.removeAnimate([index], (items) {
      final List<_Item<FormeImage>> _items =
          items.where((element) => element.value != null).toList();
      if (widget.maximum != null || widget.maximum! > items.length) {
        gridController.value = _items..add(_Item.empty());
      } else {
        gridController.value = _items;
      }
    });
  }

  void _insertPickedImages(List<FormeImage> images) {
    if (!mounted) {
      return;
    }
    final List<FormeImage> currentValue = gridController.value
        .where((element) => element.value != null)
        .map((e) => e.value!)
        .toList();
    final int num = currentValue.length;
    if (widget.maximum != null && num >= widget.maximum!) {
      return;
    }
    final Iterable<FormeImage> needInserts =
        images.where((element) => !currentValue.contains(element)).toList();
    if (needInserts.isEmpty) {
      return;
    }
    final List<FormeImage> list = [];
    for (final FormeImage item in needInserts) {
      if (!list.contains(item)) {
        list.add(item);
      }
    }

    final List<_Item<FormeImage>> items =
        (currentValue..addAll(list)).map((e) => _Item(e)).toList();

    if (widget.maximum == null) {
      gridController.value = items..add(_Item.empty());
      return;
    }

    if (items.length >= widget.maximum!) {
      gridController.value = items.sublist(0, widget.maximum);
    } else {
      gridController.value = items..add(_Item.empty());
    }
  }

  Widget _defaultGridItem(
      ValueChanged<int> remove, FormeImage item, int index) {
    Widget image;
    if (item.isBroken == false) {
      image = Image(
        image: item.cache!,
        fit: BoxFit.cover,
        width: double.infinity,
        height: double.infinity,
      );
    } else {
      image = FutureBuilder<ImageProvider>(
        future: _load(item),
        builder: (context, builder) {
          if (builder.hasData) {
            return Image(
              image: builder.data!,
              fit: BoxFit.cover,
              width: double.infinity,
              height: double.infinity,
            );
          }

          if (builder.hasError) {
            return const Center(
                child: Icon(
              Icons.broken_image,
              size: 36,
            ));
          }
          return const SizedBox.shrink();
        },
      );
    }

    return Stack(
      children: [
        Padding(
          padding: const EdgeInsets.only(
            top: 12,
            right: 12,
          ),
          child: image,
        ),
        Positioned(
          top: 0,
          right: 0,
          child: InkWell(
            highlightColor: Colors.transparent,
            splashFactory: NoSplash.splashFactory,
            hoverColor: Colors.transparent,
            focusColor: Colors.transparent,
            child: const Icon(
              Icons.cancel,
              color: Colors.red,
            ),
            onTap: () {
              remove(index);
            },
          ),
        )
      ],
    );
  }

  Widget _defaultImagePicker(
      ValueChanged<List<FormeImage>> insertPickedImages) {
    return InkWell(
      highlightColor: Colors.transparent,
      splashFactory: NoSplash.splashFactory,
      hoverColor: Colors.transparent,
      focusColor: Colors.transparent,
      onTap: () async {
        final ImagePicker _picker = ImagePicker();
        List<XFile> images;
        if (widget.maximum == 1 || controller.value.length == widget.maximum) {
          //can select only one file
          final XFile? image =
              await _picker.pickImage(source: ImageSource.gallery);
          images = image == null ? [] : [image];
        } else {
          images = await _picker.pickMultiImage() ?? [];
        }
        insertPickedImages(images.map(FormeImage.fromXFile).toList());
      },
      child: Padding(
        padding: const EdgeInsets.only(
          top: 12,
          right: 12,
        ),
        child: Container(
          color: Colors.grey,
          child: const Center(child: Icon(Icons.add)),
        ),
      ),
    );
  }

  DraggableConfiguration<_Item<FormeImage>> _draggableConfiguration() {
    if (widget.draggableConfiguration == null) {
      return const DraggableConfiguration.disabled();
    }
    return DraggableConfiguration<_Item<FormeImage>>(
      enableChecker: (item, index) {
        if (item.value == null) {
          return false;
        }
        return widget.draggableConfiguration!.enableChecker(item.value!, index);
      },
      longPressDelay: widget.draggableConfiguration!.longPressDelay,
      childWhenDraggingBuilder:
          widget.draggableConfiguration!.childWhenDraggingBuilder == null
              ? null
              : (context, item, index, child) {
                  return widget
                          .draggableConfiguration!.childWhenDraggingBuilder!(
                      context, item.value!, index, child);
                },
      feedbackBuilder: widget.draggableConfiguration!.feedbackBuilder == null
          ? null
          : (context, item, index, child, size) {
              return widget.draggableConfiguration!.feedbackBuilder!(
                  context, item.value!, index, child, size);
            },
    );
  }

  void _onAccept(List<_Item<FormeImage>>? value) {
    if (value != null) {
      gridController.value = value;
    }
  }
}

class FormeImagePickerFieldController
    extends FormeFieldControllerDelegate<List<FormeImage>> {
  FormeImagePickerFieldController(
      FormeFieldController<List<FormeImage>> delegate)
      : super(delegate);
}

abstract class FormeImage {
  Future<ImageProvider> get image;
  bool? _broken;
  ImageProvider? _cache;

  ///  typically this `image` is not a real image or can not be decoded by current platform
  ///
  /// return null if image not decoded
  bool? get isBroken => _broken;

  /// cached imageprovider after successful load from [image]
  ///
  /// will not be null if [isBroken] == false
  ImageProvider? get cache => _cache;

  /// get file name
  String? get name;

  /// get image bytes,used to save to a file
  ///
  /// there's no need to implementing this method in your custom `FormeImage`
  Future<Uint8List?> readAsBytes();

  static FormeImage fromXFile(XFile xFile) {
    return _FormeXFile(xFile);
  }

  static FormeImage fromNetwork(String url) {
    return _NetworkImage(url);
  }
}

class _NetworkImage extends FormeImage {
  final String url;

  _NetworkImage(this.url);

  @override
  Future<ImageProvider> get image async {
    return Image.network(url).image;
  }

  @override
  String? get name => throw UnimplementedError();

  @override
  Future<Uint8List?> readAsBytes() {
    throw UnimplementedError();
  }
}

class _FormeXFile extends FormeImage {
  final XFile xFile;

  _FormeXFile(this.xFile);
  @override
  Future<ImageProvider> get image async {
    return Image.memory(await readAsBytes()).image;
  }

  /// [ImagePicker] does not support get original file name in some os,but in web it does
  @override
  String? get name => xFile.name;

  @override
  Future<Uint8List> readAsBytes() async => xFile.readAsBytes();
}

class _Item<T> {
  final T? value;

  const _Item(this.value);

  _Item.empty() : value = null;
}
